@@ -9,11 +9,12 @@ Dev notes:
 
 import time
 from dataclasses import fields
+from pathlib import Path
 from typing import Optional, Tuple
 
 import numpy as np
 
-from . import corrpts, optimization, pointcloud
+from . import corrpts, mathutils, optimization, pointcloud
 
 # TODO Rename rbp (rigid-body parameters) to rbtp (rigid-body transformation parameters)
 
@@ -55,6 +56,7 @@ class SimpleICP:
         distance_weights: Optional[float] = 1,  # can also be None
         rbp_observed_values: Tuple[float] = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
         rbp_observation_weights: Tuple[float] = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
+        debug_dirpath: str = "",
     ) -> Tuple[np.array, np.array, optimization.RigidBodyParameters]:
         """Run simpleICP algorithm.
 
@@ -79,14 +81,18 @@ class SimpleICP:
                 only sense if the rigid-body transformation parameters are observed, see the
                 following arguments. Defaults to 1.
             rbp_observed_values (Tuple[float], optional): Values of direct observation of the rigid
-                body transformation parameters. Order is alpha1, alpha2, alpha3, tx, ty, tz. Unit of
+                body transformation parameters. These values define the initial transform between
+                the two point clouds. Order is alpha1, alpha2, alpha3, tx, ty, tz. Unit of
                 alpha1/2/3: degree. Defaults to (0.0, 0.0, 0.0, 0.0, 0.0, 0.0).
             rbp_observation_weights (Tuple[float], optional): Weight factors with which the
                 residuals of the direct observations are multiplied. The residuals are defined as
                 difference between the estimated rbp and the observed rbp. If an observation weight
                 is set to np.inf, this parameter is fixed to the observed value. Order of elements
                 is the same as in rbp_observed_values. Defaults to (0.0, 0.0, 0.0, 0.0, 0.0, 0.0).
-
+            debug_dirpath (str, optional): Path to directory for saving debug files, e.g. point
+                clouds and correspondences of each iteration. The xyz files can conveniently be
+                inspected with CloudCompare. Not considered if an empty string is passed. Defaults
+                to "".
         Returns:
             Tuple[np.array, np.array]:
                 H: Estimated homogeneous transformation matrix.
@@ -100,15 +106,29 @@ class SimpleICP:
 
         start_time = time.time()
 
+        if debug_dirpath:
+            print(f'Write debug files to directory "{debug_dirpath}"')
+            Path(debug_dirpath).mkdir(parents=True, exist_ok=True)
+
         # Convert angle valus from degree -> radian
         rbp_observed_values = np.array(rbp_observed_values)
         for i in range(3):
             rbp_observed_values[i] = rbp_observed_values[i] * np.pi / 180
 
+        # Compose initial H from observed rbp values
+        H = mathutils.create_homogeneous_transformation_matrix(
+            mathutils.euler_angles_to_rotation_matrix(
+                rbp_observed_values[0], rbp_observed_values[1], rbp_observed_values[2]
+            ),
+            rbp_observed_values[3:],
+        )
+
         if np.isfinite(max_overlap_distance):
 
             print("Consider partial overlap of point clouds ...")
+            self.pc2.transform_by_H(H)  # temporarily transform pc2
             self.pc1.select_in_range(self.pc2.X, max_range=max_overlap_distance)
+            self.pc2.transform_by_H(np.linalg.inv(H))  # undo transformation
 
             if not self.pc1.num_selected_points > 0:
                 raise SimpleICPException(
@@ -125,8 +145,6 @@ class SimpleICP:
             print("Estimate normals of selected points ...")
             self.pc1.estimate_normals(neighbors)
 
-        H = np.eye(4)
-
         distance_residuals = []
 
         print("Start iterations ...")
@@ -134,16 +152,29 @@ class SimpleICP:
 
             cp = corrpts.CorrPts(self.pc1, self.pc2)
 
-            # We need to temporarily transform the moving point cloud for matching
-            self.pc2.transform_by_H(H)
+            self.pc2.transform_by_H(H)  # temporarily transform pc2
+            if debug_dirpath:
+                self.pc1.write_xyz(
+                    Path(debug_dirpath).joinpath(f"iteration{it:03d}_pcfix.xyz")
+                )
+                self.pc2.write_xyz(
+                    Path(debug_dirpath).joinpath(f"iteration{it:03d}_pcmov.xyz")
+                )
             cp.match()
-            self.pc2.transform_by_H(np.linalg.inv(H))
+            self.pc2.transform_by_H(np.linalg.inv(H))  # undo transformation
 
             # Rejection of possibly false correspondences
             cp.reject_wrt_planarity(min_planarity)
             cp.reject_wrt_point_to_plane_distances()
             # cp.reject_wrt_to_angle_between_normals()  # not implemented yet
 
+            if debug_dirpath:
+                cp.write_xyz(
+                    Path(debug_dirpath).joinpath(
+                        f"iteration{it:03d}_correspondences.xyz"
+                    )
+                )
+
             if it == 0:
                 initial_distances = cp.point_to_plane_distances
                 rbp_initial_values = rbp_observed_values
